#pragma once
#include <boost/interprocess/sync/interprocess_mutex.hpp>
#include <boost/interprocess/sync/interprocess_condition.hpp>
#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/containers/map.hpp>
#include <boost/interprocess/containers/vector.hpp>
#include <boost/interprocess/containers/string.hpp>
#include <boost/interprocess/allocators/allocator.hpp>
#include <boost/interprocess/smart_ptr/unique_ptr.hpp>
#include <boost/interprocess/smart_ptr/deleter.hpp>
#include <boost/cstdint.hpp>

#include <string>

struct CommandShm;
struct CommandInterface;
typedef boost::shared_ptr<CommandInterface> CommandInterfaceSPtr;

/*! @brief a handle for a shared memory communication struct
 *
 */
struct CommandInterface {

	//! each segment has its own name after it was created.
	//! The name is a combination of what's given in the constructor
	//! as a service name and the static suffix of CommandShm
	std::string                                  m_segment_name;
	std::string                                  m_command_struct_name;

	CommandShm                                  *m_command_shm;

	// I have to store the segment object as it destroys 
	// everything in it when it gets destroyed itself
	// note that this is not an offset_ptr as it lives in local
	// not interprocess memory   -----  or so he thought.
	boost::interprocess::managed_shared_memory  *m_command_segment;
};

typedef boost::interprocess::allocator<char, boost::interprocess::managed_shared_memory::segment_manager> ShmemCharAllocator;
typedef boost::interprocess::basic_string<char, std::char_traits<char>, ShmemCharAllocator> ShmemString;
typedef boost::interprocess::allocator<ShmemString, boost::interprocess::managed_shared_memory::segment_manager> ShmemStringAllocator;
typedef boost::interprocess::deleter<ShmemString, boost::interprocess::managed_shared_memory::segment_manager> ShmemStringDeleter;

typedef boost::interprocess::managed_unique_ptr<ShmemString, boost::interprocess::managed_shared_memory>::type ShmemStringUPtr;
typedef boost::interprocess::offset_ptr<ShmemString> ShmemStringPtr;


typedef std::pair<const boost::uint64_t, ShmemStringPtr>  MessageValueType;
typedef boost::interprocess::allocator<MessageValueType, boost::interprocess::managed_shared_memory::segment_manager>  ShmemAllocator;
typedef boost::interprocess::map<boost::uint64_t, ShmemStringPtr, std::less<boost::uint64_t>, ShmemAllocator> MessageMap;



inline void msgmap_insert(MessageMap *n_map, CommandInterfaceSPtr n_interface,
									const boost::uint64_t n_key, const std::string &n_value) {
	
	boost::interprocess::managed_shared_memory::segment_manager *seg_mngr = n_interface->m_command_segment->get_segment_manager();

	// create the string
	ShmemStringPtr  s = seg_mngr->construct<ShmemString>(boost::interprocess::anonymous_instance)(n_value.c_str(), seg_mngr);
	(*n_map)[n_key] = s;
}



/*! @brief Shared Memory command segment
*/ 
struct CommandShm {

	public:
		static const char *s_segment_suffix;          //!< suffix name of the shared segment
		static const char *s_command_struct_suffix;   //!< suffix name of this struct in segment

		CommandShm();
		virtual ~CommandShm() noexcept;

		/*! @brief delete potentially existing segment, create and setup a new one 
		 *   with the given service name
		 *  \param n_service_name must be ASCII127, short and not null
		 *  \throw bad_alloc on fail
		 */
		static CommandInterfaceSPtr create(const char *n_service_name);

		/*! @brief open and setup for use
		 *  \param n_service_name must be ASCII127, short and not null
		 *  \throw nil
		 *  \return nullptr on fail
		 */
		static CommandInterfaceSPtr open(const char *n_service_name) noexcept;

		enum { total_size = 512 * 1024 };   //!< 512k should be enough for that one here

	

		// Mutex to protect access to the the command map
		boost::interprocess::interprocess_mutex      m_command_mutex;

		// when this turns true, a command is available
		boost::interprocess::interprocess_condition  m_command_condition;

		// put your command string in here with a random command ID
		boost::interprocess::offset_ptr<MessageMap>  m_commands;

		// Mutex to protect access to the the response map
		boost::interprocess::interprocess_mutex      m_response_mutex;

		// when this notifies, a new response is available
		boost::interprocess::interprocess_condition  m_response_condition;

		// get your response here with your command ID
		boost::interprocess::offset_ptr<MessageMap>  m_responses;
};
