//  Copyright 2016 (c) Stephan Menzel, vr-on GmbH

#include "ShmServer.hpp"

#include <boost/cstdint.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#include <boost/date_time/posix_time/ptime.hpp>

#include <exception>

using namespace boost::interprocess;

ShmServer::ShmServer()
		: m_command_thread()
		, m_interrupted(false)
		, m_command_memory(CommandShm::create("test")) {

	if (!m_command_memory) {
		throw std::bad_alloc();
	}
	
	assert(m_command_memory);
}

ShmServer::~ShmServer() noexcept {

	if (m_command_memory && m_command_memory->m_command_segment) {
		delete m_command_memory->m_command_segment;
		m_command_memory->m_command_segment = nullptr;
	};

	shared_memory_object::remove(m_command_memory->m_segment_name.c_str());
}

void ShmServer::start() {

	// start a thread that will handle listening on our command buffer
	assert(!m_command_thread);
	m_command_thread.reset(new boost::thread([&]() { this->command_thread(); }));
}

void ShmServer::stop() noexcept {

	if (m_command_thread) {
		m_interrupted.store(1);
		m_command_thread->interrupt();  // sadly, interprocess condition won't react on this interrupt
		m_command_thread->join();
		m_command_thread.reset();
	}
}

// this loops over the shm command buffer 
void ShmServer::command_thread() noexcept {
	
	try {
		// I copy the commands out of the buffer to release 
		// the lock early and then process
		std::map<boost::uint64_t, std::string> commands;
		std::map<boost::uint64_t, std::string> responses;

		CommandShm *cbuffer = m_command_memory->m_command_shm;

		while (!m_interrupted.load()) {
			{
				scoped_lock<interprocess_mutex> lock(cbuffer->m_command_mutex);
				if (cbuffer->m_commands->empty()) {
					cbuffer->m_command_condition.timed_wait(lock, boost::get_system_time() + boost::posix_time::milliseconds(100));
				} else {
					for (const MessageMap::value_type &v : *(cbuffer->m_commands)) {
						commands[v.first] = std::string(v.second->cbegin(), v.second->cend());   // I know, there should be a more elegant std::copy but I can't be arsed

						// now for clearing the map, I have to delete the string by hand
						// as I seem to be too stupid to use interprocess' smart ptrs
						m_command_memory->m_command_segment->get_segment_manager()->destroy_ptr(v.second.get());
					}

					cbuffer->m_commands->clear();
				}
			}

			// now the lock is dropped, process commands
			for (std::pair<const boost::uint64_t, std::string> &c : commands) {

				// This is where user server code takes over and returns stuff. I'll fill in a string response.
				responses[c.first] = std::string("Hello World Response");
			}

			commands.clear();

			// now we're done, copy responses into the control buffer. 
			// I will not delete potentially existing responses that haven't been consumed
			{
				scoped_lock<interprocess_mutex> lock(cbuffer->m_response_mutex);
				for (std::pair<const boost::uint64_t, std::string> &r : responses) {
					msgmap_insert(cbuffer->m_responses.get(), m_command_memory, r.first, r.second);
				}

				cbuffer->m_response_condition.notify_all();
			}

			responses.clear();

			// rinse repeat
		}
	} catch (const std::exception &sex) {
		std::cerr << "Exception caught in shm loop: " << boost::diagnostic_information(sex) << std::endl;
	}
}
